//----------------------------------------------------------------------------------------------------------------------
#include "../io/Console.hpp"
#include "../tex/Texture.hpp"
#include "../debug.hpp"
#include "Effect.hpp"
#include "EffectManager.hpp"
//----------------------------------------------------------------------------------------------------------------------
using namespace bleifrei::io;
using namespace bleifrei::math;
using namespace bleifrei::tex;
using namespace std;
//----------------------------------------------------------------------------------------------------------------------
namespace bleifrei {
	namespace gfx {
//----------------------------------------------------------------------------------------------------------------------

bool
Effect::setTechnique(Technique &technique)
{
	currentTechnique = &technique;
	return cgIsTechniqueValidated(technique.m_cgTechnique) == CG_TRUE;
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::setParameter(Parameter &parameter, const float value)
{
	cgSetParameter1f(parameter.m_cgParameter, value);
	return true;
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::setParameter(Parameter &parameter, const Vector3f &value)
{
	cgSetParameter3fv(parameter.m_cgParameter, &value[0]);
	return true;
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::setParameter(Parameter &parameter, const Vector4f &value)
{
	cgSetParameter4fv(parameter.m_cgParameter, &value[0]);
	return true;
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::setParameter(Parameter &parameter, const Matrix4x4f &matrix)
{
	cgGLSetMatrixParameterfr(parameter.m_cgParameter, &matrix[0]);
	return true;
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::setParameter(Parameter &parameter, Texture *texture)
{
	if (texture)
	{
		cgGLSetupSampler(parameter.m_cgParameter, texture->opengl_getTextureID());
	}
	return true;
}

//----------------------------------------------------------------------------------------------------------------------

Effect::technique_iterator
Effect::getTechniqueByName(const string &name)
{
	technique_iterator tit = m_techniques.begin();
	while (tit != m_techniques.end() && tit->getName() != name)
	{
		++tit;
	}
	return tit;
}

//----------------------------------------------------------------------------------------------------------------------

Effect::parameter_iterator
Effect::getParameterByName(const string &name)
{
	parameter_iterator pit = m_parameters.begin();
	while (pit != m_parameters.end() && pit->getName() != name)
	{
		++pit;
	}
	return pit;
}

//----------------------------------------------------------------------------------------------------------------------

Effect::parameter_iterator
Effect::getParameterBySemantic(const string &semantic)
{
	parameter_iterator pit = m_parameters.begin();
	while (pit != m_parameters.end() && pit->getSemantic() != semantic)
	{
		++pit;
	}
	return pit;
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::begin(unsigned int &passes)
{
	if (!currentTechnique)
	{
		return false;
	}
	passes = (unsigned int)currentTechnique->m_passes.size();
	return true;
}

//----------------------------------------------------------------------------------------------------------------------

void
Effect::end(void)
{
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::beginPass(unsigned int pass)
{
	if (!currentTechnique)
	{
		return false;
	}
	m_currentPass = pass;
	cgSetPassState(currentTechnique->m_passes[m_currentPass].m_cgPass);
	return true;
}

//----------------------------------------------------------------------------------------------------------------------

void
Effect::endPass()
{
	if (currentTechnique)
	{
		cgResetPassState(currentTechnique->m_passes[m_currentPass].m_cgPass);
	}
}

//----------------------------------------------------------------------------------------------------------------------

Effect::Effect(void) :
	m_cgEffect(0),
	currentTechnique(0)
{
}

//----------------------------------------------------------------------------------------------------------------------

Effect::~Effect(void)
{
	unload();
}

//----------------------------------------------------------------------------------------------------------------------

bool
Effect::load(const string &filename)
{
	unload();

	CGcontext cgContext = EffectManager::getInstance().cg_getContext();
	const char *errors = 0;
	m_cgEffect = cgCreateEffectFromFile(cgContext, filename.c_str(), 0);
	
	if (!m_cgEffect)
	{
		string errorMessage = cgGetLastListing(cgContext);
		string line;
	
		Console::getInstance().printf(1, 0, 0, "ERROR(%s): Failed to load effect: %s", "Compile Error", filename.c_str());

		string::iterator it;
            for (it = errorMessage.begin(); it != errorMessage.end(); ++ it)
		{
			if (*it == '\n')
			{
				Console::getInstance().printf(1, 0, 0, "    %s", line.c_str());
				line = "";
			}
			else
			{
				line += *it;
			}
		}
		Console::getInstance().printf(1, 0, 0, "    %s", line.c_str());
		return false;
	}

	for (CGtechnique cgTechnique = cgGetFirstTechnique(m_cgEffect); cgTechnique; cgTechnique = cgGetNextTechnique(cgTechnique))
	{
		if (cgValidateTechnique(cgTechnique) == CG_FALSE)
		{
			// $TODO: check for error
			continue;
		}
		
		m_techniques.push_back(Technique(cgTechnique));
		Technique& technique = m_techniques.back();
		
		for (CGpass cgPass = cgGetFirstPass(cgTechnique); cgPass; cgPass = cgGetNextPass(cgPass))
		{
			technique.m_passes.push_back(Pass(cgPass));
		}
	}

	for (CGparameter cgParameter = cgGetFirstEffectParameter(m_cgEffect); cgParameter; cgParameter = cgGetNextParameter(cgParameter))
	{
		m_parameters.push_back(Parameter(cgParameter));
	}

	return true;
}

//----------------------------------------------------------------------------------------------------------------------

void
Effect::unload(void)
{
	currentTechnique = 0;
	if (m_cgEffect)
	{
		m_techniques.clear();
		cgDestroyEffect(m_cgEffect);
		m_cgEffect = 0;
	}
}

//----------------------------------------------------------------------------------------------------------------------
	}
}
//----------------------------------------------------------------------------------------------------------------------
