
// includes
#include <memory>

#include "../io/io.hpp"
#include "../debug.hpp"

#include "Image.hpp"

using namespace bleifrei::io;
using namespace bleifrei::platform;
using namespace std;

namespace bleifrei {
	namespace img {


		#pragma pack(push, 1)

		// Image header
		struct ImageHeader
		{
			word width;
			word height;
			byte format;
			byte palette;
			byte compression;
			byte reserved;
		};

		#pragma pack(pop)

		inline static word rgb16bit555(byte r, byte g, byte b)
		{
			return (b >> 3) + ((g & 0xF8) << 2) + ((r & 0xF8) << 7);
		}

		inline static word rgb16bit565(byte r, byte g, byte b)
		{
			return (b >> 3) + ((g & 0xF8) << 3) + ((r & 0xF8) << 8);
		}

		/** Returns the rgb components of the pixel at src.
		 * The color components of the pixel pointed to by \c src are extracted and converted
		 * into byte values. \c format tells the function how to decode the pixel data.
		 * @param "byte *src" byte pointer to source pixel
		 * @param "ImageFormat format" image format
		 * @param "byte &r" Reference to byte where the red component is to be stored
		 * @param "byte &g" Green component
		 * @param "byte &b" Blue component
		 * @param "byte &alpha" Alpha component
		 * @returns \c true if the format is valid, \c false otherwise
		 */
		inline static bool
		getRGB(byte *src, ImageFormat format, byte &r, byte &g, byte &b, byte &alpha)
		{
			word v;

			switch (format) {
			case ImageFormatA8 :
				alpha = *src;
				break;
			case ImageFormatL8 :
				r = g = b = *src;
			case ImageFormatR5G5B5A1 :
				v = *((unsigned short *)src);
				r = byte((v & 0x7C00) >> 10);
				g = byte((v & 0x03E0) >> 5);
				b = byte(v & 0x001F);
				break;
			case ImageFormatB5G5R5A1 :
				v = *((unsigned short *)src);
				r = byte((v & 0x7C00) >> 10);
				g = byte((v & 0x03E0) >> 5);
				b = byte(v & 0x001F);
				break;
			case ImageFormatR5G6B5 :
				v = *((unsigned short *)src);
				r = byte((v & 0xF800) >> 11);
				g = byte((v & 0x07E0) >> 5);
				b = byte(v & 0x001F);
				break;
			case ImageFormatB5G6R5 :
				v = *((unsigned short *)src);
				r = byte((v & 0xF800) >> 11);
				g = byte((v & 0x07E0) >> 5);
				b = byte(v & 0x001F);
				break;
			case ImageFormatR8G8B8A8 :
				alpha = src[3];
			case ImageFormatR8G8B8 :
				r = src[2];
				g = src[1];
				b = src[0];
				break;
			case ImageFormatB8G8R8A8 :
				alpha = src[3];
			case ImageFormatB8G8R8 :
				b = src[2];
				g = src[1];
				r = src[0];
				break;
			default:
				BLEIFREI_ERROR("Unknown image format");
				return false;
				break;
			}
			return true;
		}

		/** Sets pixel at dest to RGBA.
		 * The arguments r, g, b and alpha are converted to the pixel format specified by \c format
		 * and stored at the location pointed to by \c dest.
		 * The color components of the pixel pointed to by \c src are extracted and converted
		 * into byte values. \c format tells the function how to decode the pixel data.
		 * @param "byte *dest" byte pointer to destination pixel
		 * @param "ImageFormat format" Image format
		 * @param "byte r" Red component
		 * @param "byte g" Green component
		 * @param "byte b" Blue component
		 * @param "byte alpha" Alpha component
		 * @returns \c true if the format is valid, \c false otherwise
		 */
		inline static bool
		putRGB(byte *dest, ImageFormat format, byte r, byte g, byte b, byte alpha)
		{

			// Convert 32 bit data to the correct format
			switch (format) {
			case ImageFormatA8 :
				*dest = alpha;
				break;
			case ImageFormatL8 :
				*dest = (r + b + g) / 3;
				break;
			case ImageFormatR5G5B5A1 :
				*((word *)dest) = rgb16bit555(r,g,b);
				break;
			case ImageFormatB5G5R5A1 :
				*((word *)dest) = rgb16bit555(b,g,r);
				break;
			case ImageFormatR5G6B5 : 
				*((word *)dest) = rgb16bit565(r,g,b);
				break;
			case ImageFormatB5G6R5 : 
				*((word *)dest) = rgb16bit565(b,g,r);
				break;
			case ImageFormatR8G8B8A8 :
				dest[3] = alpha;
			case ImageFormatR8G8B8 :
				// 24 bits
				dest[2] = r;
				dest[1] = g;
				dest[0] = b;
				break;
			case ImageFormatB8G8R8A8 :
				dest[3] = alpha;
			case ImageFormatB8G8R8 :
				// 24 bits
				dest[2] = b;
				dest[1] = g;
				dest[0] = r;
				break;
			default:
				BLEIFREI_ERROR("Unknown image format");
				return false;
			}
			return true;
		}

		/** Create new empty image.
		 * @param none
		 */
		Image::Image(void) :
			m_imageData(0)
		{
		}

		/** Create image and read it from stream.
		 * @param "istream &is" Source stream
		 */
		Image::Image(istream &is) :
			m_imageData(0)
		{
			load(is);
		}

		/** Create new empty image with specified bit depth.
		 * @param "ImageFormat format" ImageFormat
		 */
		Image::Image(ImageFormat format) :
			m_imageData(0),
			m_format(format)
		{
		}

		/** Create a new image and initialize it to the specified dimensions.
		 * @param "int width" Width of the image
		 * @param "int height" Width of the image
		 * @param "ImageFormat format" Format fo image
		 */
		Image::Image(int width, int height, ImageFormat format) :
			m_imageData(0)
		{
			create(width, height, format);
		}

		/** Create new image and red RAW data from a stream.
		 * The parameters width, height and bpp are used to create a new image of the
		 * specified dimensions and to calculate how much data should be read from the stream.
		 * read raw image data from the stream.
		 * @param "istream &is" Source stream
		 * @param "int width" Width of the image
		 * @param "int height" Width of the image
		 * @param "ImageFormat format" Format fo image
		 * @throws stream_exception
		 */
		Image::Image(istream &is, int width, int height, ImageFormat format) :
			m_imageData(0)
		{
			// Create new image
			create(width, height, format);

			// Read image data
			is.read(((char *)(getData())), getBytesPerLine() * getHeight());

			// If error throw exception
			if (!is) throw stream_exception();
		}

		/** Initialize image data.
		 * Reset the Image to the specified dimensions.
		 * @param "int width" Width of the image
		 * @param "int height" Width of the image
		 * @param "ImageFormat format" Bit depth fo image
		 * @throws out_of_memory_exception
		 * @returns void
		 */
		void
		Image::create(int width, int height, ImageFormat format)
		{
			destroy();

			m_width = width;
			m_height = height;
			m_format = format;

			// Calculate unsupplied attributes
			m_pixelwidth = m_pixelwidth = (imageFormatToBitDepth(format) + 7) / 8;			// Width in bytes
			m_bytesPerLine = m_pixelwidth * m_width;

			// Allocate memory
			m_imageData = new byte[m_bytesPerLine * m_height];
			if (!m_imageData)
			{
				throw out_of_memory_exception();
			}
		}

		/** Destroys the image.
		 */
		Image::~Image(void)
		{
			destroy();
		}

		/** Destroys the image.
		 * Deallocates memory
		 * @param none
		 * @returns void
		 */
		void Image::destroy(void)
		{
			if (m_imageData)
			{
				delete[] m_imageData;
			}
			m_imageData = 0;
		}

		/** Flips the bitmap upside down.
		 * @param none
		 * @throws out_of_memory_exception
		 * @returns void
		 */
		void
		Image::flip(void)
		{
			// Create temporary storage for one line
			byte *temp = new byte[getBytesPerLine()];

			// throw Exception if not enough memory
			if (!temp)
			{
				throw out_of_memory_exception();
			}

			// Flip lines one by one
			for (int index = 0; index < getHeight() / 2; index++)
			{
				// First line -> temp
				memcpy(temp, &getData()[index * getBytesPerLine()], getBytesPerLine());

				// Last line -> First line
				memcpy(&getData()[index * getBytesPerLine()], &getData()[(getHeight() - 1 - index) * getBytesPerLine()], getBytesPerLine());

				// Temp -> Last line
				memcpy(&getData()[(getHeight() - 1 - index) * getBytesPerLine()], temp, getBytesPerLine());
			}

			// Free temporary memory
			delete[] temp;
		}

		/** Sets pixel at (x,y) to color (r, g, b, alpha)
		 * Converts the color components to the correct pixelformat and saves it at
		 * position (x, y).
		 * @param "int x" X location
		 * @param "int y" Y location
		 * @param "unsigned char r" Red component of color
		 * @param "unsigned char g" Green component of color
		 * @param "unsigned char b" Blue component of color
		 * @param "unsigned char alpha" Alpha component of color (if 32bit)
		 * @returns void
		 */
		void
		Image::putPixel(int x, int y, byte r, byte g, byte b, byte alpha)
		{
			byte *dest = &getData()[x * getPixelWidth() + getBytesPerLine() * y];

			putRGB(dest, getFormat(), r, g, b, alpha);
		}

		/** Return RGBA from pixel at (x,y).
		 * Converts the color components from the pixel at position (x,y) to bytes and sets the
		 * values of the references r, g, b and alpha.
		 * @param "int x" X location
		 * @param "int y" Y location
		 * @param "unsigned char &r" Reference to byte where to store red component of color
		 * @param "unsigned char &g" Green component of color
		 * @param "unsigned char &b" Blue component of color
		 * @param "unsigned char alpha" Alpha component of color (if 32bit)
		 * @returns void
		 */
		void Image::getPixel(int x, int y, byte &r, byte &g, byte &b, byte &alpha) {
			
			// Set source pointer
			byte *source = &getData()[y * getBytesPerLine() + x * getPixelWidth()];

			getRGB(source, getFormat(), r, g, b, alpha);
		}

		/** Copy image data from one Image to another.
		 * Uses the methods getPixel and putPixel to copy the pixels from one image to another.
		 * @warning Completely destroys old image data
		 * @param int bpp Bit depth of destination image
		 * @throws out_of_memory_exception
		 * @returns void
		 * @author @top
		 */
		void Image::copy(Image &image, ImageFormat format) {
			byte r, g, b, alpha;

			create(image.getWidth(), image.getHeight(), format);

			// Copy each pixel (Slow)
			for (int y = 0; y < getHeight(); y++) {
				for (int x = 0; x < getWidth(); x++) {
					r = g = b = alpha = 0;
					image.getPixel(x, y, r, g, b, alpha);
					putPixel(x, y, r, g, b, alpha);
				}
			}
		}

		/** Decodes Run-Length Encoded data from a stream.
		 * @param "istream &is" Source stream
		 * @param "int numbytes" Size of RLE data in bytes
		 * @throws stream_exception
		 * @returns void
		 */
		void Image::decodeRLE(istream &is, int numbytes) {
			byte c;

			// While there are still bytes left to read from the stream
			for (int ptr = 0; ptr < numbytes;) {

				if (!is) throw stream_exception();			// Is the stream OK?

				c = is.get();								// Get next byte
				if (c & 0x80) {								// Is it a 'run'?
					byte color[4];
	
					c &= 0x7f;								// Get pixel count
					c++;

					is.read((char *)color, getPixelWidth());		// Read color from stream
					
					for (int i = 0; i < c; i++) {			// Write color c times
						for (int j = 0; j < getPixelWidth(); j++, ptr++) {
							getData()[ptr] = color[j];
						}
					}
				}
				else {										// Read count pixels
					c++;
					is.read((char *)&getData()[ptr], c * getPixelWidth());
					ptr += c * getPixelWidth();
				}
			}
		}

		/** Compare two pixels (Used when encoding RLE data).
		 * @param "byte *p" Pointer to pixel data
		 * @param "int pixelwidth" Width of a pixel in bytes
		 * @returns \c true if the pixels match, \c false otherwise
		 */
		inline static bool comparepixels(byte *p, int pixelwidth) {

			// Test all bytes of pixel
			for (int i = 0; i < pixelwidth; i++) {

				// Does pixel match the next one? If not return false
				if (p[i] != p[i + pixelwidth]) return false;
			}
			return true;
		}

		/** Write a 'run' to a stream (RLE help function).
		 * @param "ostream &os" Destination stream
		 * @param "byte *src" Source pixel
		 * @param "int count" Number of pixels to write
		 * @param "int pixelwidth" Width of a pixel in bytes
		 * @returns void
		 */
		inline static void writeRLE(ostream &os, byte *src, int count, int pixelwidth) {
			os.put((byte)((count - 1) | 0x80));
			os.write((char *)src, pixelwidth);
		}

		/** Copy count pixels from memory to a stream (RLE help function).
		 * @param "ostream &os" Destination stream
		 * @param "byte *src" Byte pointer to source pixels
		 * @param "int count" Number of pixels to write
		 * @param "int pixelwidth" Width of a pixel in bytes
		 * @returns void
		 */	
		inline static void writeRAW(ostream &os, byte *src, int count, int pixelwidth) {
			os.put((byte)(count - 1));
			os.write((char *)src, count * pixelwidth);
		}

		/** Writes RLE data to a stream.
		 * @param "ostream &os" Destination stream
		 * @returns void
		 */
		void Image::encodeRLE(ostream &os) {

			// Pointer to the first pixel to copy
			byte *firstpixel = NULL;

			// Number of pixels in current 'run'
			int c1 = 0;

			// Number of pixels to copy
			int c2 = 0;

			int ptr = 0;
			int size = getBytesPerLine() * getHeight() - getPixelWidth();

			// While there are still pixels to write
			while (ptr < size) {
				int isequal = comparepixels(&getData()[ptr], getPixelWidth());

				// If next pixel equals the last ones and there are less than
				// 127 pixels in this repeat operation
				if (isequal && c1 < 127) {

					// Increase "run" length and pointer
					c1++;
					ptr += getPixelWidth();

					// If c2 > 0 there are c2 pixels from a previous copy
					// operation to write to stream
					if (c2) {
						
						// Write any pixels
						writeRAW(os, firstpixel, c2, getPixelWidth());

						// Reset copy counter and pointer to first pixel
						firstpixel = NULL;
						c2 = 0;
					}
				}
				// If there are 127 pixels write them to the stream
				else if (c1) {
						writeRLE(os, &getData()[ptr], c1 + 1, getPixelWidth());
						c1 = 0;
						ptr += getPixelWidth();
				}
				// 
				else {

					// Start new run
					if (firstpixel == NULL) firstpixel = &getData()[ptr];

					// Update pointers
					c2++;
					ptr += getPixelWidth();

					// Copy 128 pixels to stream
					if (c2 == 128) {
						writeRAW(os, firstpixel, c2, getPixelWidth());

						// Reset pointer and "run" length
						firstpixel = NULL;
						c2 = 0;
					}
				}
			}

			// If there are still unwritten pixels when we exit the loop we
			// write them
			if (c1) {
				writeRLE(os, &getData()[ptr], c1 + 1, getPixelWidth());
			}
			else if (c2) {
				writeRAW(os, firstpixel, c2 + 1, getPixelWidth());
			}
		}

		/** Writes an image to an ostream, compressed or uncompressed.
		 * @param "ImageCompression compression" Compression algorithm
		 * @returns void
		 */
		void Image::save(ostream &os, ImageCompression compression) {
			FileId id;
			ImageHeader header;

			if (compression > ImageCompressionRLE || compression < ImageCompressionRaw) {
				compression = ImageCompressionRaw;			// If unknown compression, force RAW
			}

			// Fill file header
			id = string2id("IMG1");

			header.height = m_height;					// Image width
			header.width = m_width;						// Image height
			header.format = m_format;					// Image bit depth
			header.palette = 0;							// Image does not have palette (Not supported!)
			header.compression = compression;			// Image compression format
			header.reserved = 0;						// Padding

			os.write((char *)&id, sizeof(id));			// Write file header

			os.write((char *)&header, sizeof(header));	// Write image header

			switch (compression) {						// Select compression
			case ImageCompressionRaw :
				os.write((char *)getData(), getBytesPerLine() * getHeight());
				break;
			case ImageCompressionRLE :
				encodeRLE(os);
				break;
			}
		}

		/** Reads an image from an istream.
		 * @param "istream &is" Source stream
		 * @throws file_format_exception, out_of_memory_exception
		 * @returns void
		 */
		void Image::load(istream &is) {
			FileId id;
			ImageHeader header;

			// Read file header
			is.read((char *)&id, sizeof(id));

			// Make sure it is an imagefile
			if (id != string2id("IMG1")) throw file_format_exception();
			
			// Read image header
			is.read((char *)&header, sizeof(header));

			// Read palette if there is one
			if (header.palette) {
				throw file_format_exception();
			}

			// Initialize image    
			create(header.width, header.height, static_cast<ImageFormat>(header.format));

			// Decode imagedata
			switch (header.compression) {
			case ImageCompressionRaw :
				is.read((char *)getData(), getBytesPerLine() * getHeight());
				break;
			case ImageCompressionRLE :
				decodeRLE(is, getBytesPerLine() * getHeight());
				break;
			default:
				throw file_format_exception();
			}
		}

		/** Convert image data from BGR to RGB (32 bit).
		 * @param none	
		 * @returns void
		 * @author @top
		 */
		void Image::convertBGRtoRGB32() {
			byte *maxPtr = getData() + getBytesPerLine() * getHeight();
			for (byte *ptr = getData(); ptr < maxPtr; ptr += 4) {
				ptr[2] ^= ptr[0];
				ptr[0] ^= ptr[2];
				ptr[2] ^= ptr[0];
			}
		}

		/** Convert image data from BGR to RGB (24 bit).
		 * @param none	
		 * @returns void
		 * @author @top
		 */
		void
		Image::convertBGRtoRGB24(void)
		{
			byte *maxPtr = getData() + getBytesPerLine() * getHeight();
			for (byte *ptr = getData(); ptr < maxPtr; ptr += 3) {
				ptr[2] ^= ptr[0];
				ptr[0] ^= ptr[2];
				ptr[2] ^= ptr[0];
			}
		}

		/** Convert image data from BGR to RGB (all formats).
		 * @param none	
		 * @returns void
		 * @author @top
		 */
		void
		Image::convertBGRtoRGB(void)
		{
			for (int i = 0; i < getHeight(); i++) {
				for (int j = 0; j < getWidth(); j++) {
					byte r, g, b, alpha;
					getPixel(j, i, r, g, b, alpha);
					putPixel(j, i, b, g, r, alpha);
				}
			}
		}

	}
}
