
//-------------------------------------------------------------------------------------------------
// carscene-1.fx
//-------------------------------------------------------------------------------------------------

//-------------------------------------------------------------------------------------------------
// Parameters with semantics and default values.
//-------------------------------------------------------------------------------------------------
//texture framebufferMap : FRAME_BUFFER_TEXTURE;
float4 windowPosition : WINDOW_POSITION = float4(0,0,1,1);

//-------------------------------------------------------------------------------------------------
// framebufferSampler
//		Texture sampler for the framebuffer texture.
//-------------------------------------------------------------------------------------------------
samplerRECT framebufferSampler : FRAME_BUFFER_TEXTURE = sampler_state
{
//    Texture = <framebufferMap>;
//    MinFilter = Linear;
//    MagFilter = Linear;
//    MipFilter = Linear;
};

samplerRECT firstframeSampler : FIRST_FRAME_TEXTURE = sampler_state
{
//    Texture = <framebufferMap>;
//    MinFilter = Linear;
//    MagFilter = Linear;
//    MipFilter = Linear;
};

//-------------------------------------------------------------------------------------------------
// Variables used for YIQ/RGB color space conversion.
//-------------------------------------------------------------------------------------------------
float3 yiqTransform0 = { 0.299, 0.587, 0.144 };
float3 yiqTransform1 = { 0.596,-0.275,-0.321 };
float3 yiqTransform2 = { 0.212,-0.523, 0.311 };

float3 yiqInverseTransform0 = { 1, 0.956, 0.621 };
float3 yiqInverseTransform1 = { 1,-0.272,-0.647 };
float3 yiqInverseTransform2 = { 1,-1.105, 1.702 };

//-------------------------------------------------------------------------------------------------
// psRed
//		Pixel shader used for modulating framebuffer texture with red.
//-------------------------------------------------------------------------------------------------
float4 psRed(float2 tex0 : TEXCOORD0, uniform samplerRECT framebufferSampler) : COLOR
{
	// sample framebuffer texture
	float4 framebuffer = texRECT(framebufferSampler, tex0);

	// modulate sample with red (1, 0, 0, 1)
	return float4(1, 0, 0, 1) * framebuffer;
}

//-------------------------------------------------------------------------------------------------
// psBlur
//		Pixel shader used for 4 sample blur filter.
//-------------------------------------------------------------------------------------------------
float4 psBlur(float2 tex0 : TEXCOORD0, uniform samplerRECT framebufferSampler) : COLOR
{
	// sample framebuffer texture at current texel
	float4 framebuffer = 0.5 * texRECT(framebufferSampler, tex0);

	// sample framebuffer texture at neighbouring texels
	framebuffer += 0.125 * texRECT(framebufferSampler, tex0 + float2( 1, 0));
	framebuffer += 0.125 * texRECT(framebufferSampler, tex0 + float2(-1, 0));
	framebuffer += 0.125 * texRECT(framebufferSampler, tex0 + float2( 0, 1));
	framebuffer += 0.125 * texRECT(framebufferSampler, tex0 + float2( 0,-1));
	
	// returned weighted sum
	return framebuffer;
}

//-------------------------------------------------------------------------------------------------
// psSharpen
//		Pixel shader used for 4 sample sharpen filter.
//-------------------------------------------------------------------------------------------------
float4 psSharpen(float2 tex0 : TEXCOORD0, uniform samplerRECT framebufferSampler) : COLOR
{
	// sample framebuffer texture at current texel
	float4 framebuffer = 1.5 * texRECT(framebufferSampler, tex0);
	
	// sample framebuffer texture at neighbouring texels
	framebuffer -= 0.125 * texRECT(framebufferSampler, tex0 + float2( 1, 0));
	framebuffer -= 0.125 * texRECT(framebufferSampler, tex0 + float2(-1, 0));
	framebuffer -= 0.125 * texRECT(framebufferSampler, tex0 + float2( 0, 1));
	framebuffer -= 0.125 * texRECT(framebufferSampler, tex0 + float2( 0,-1));
	
	// returned weighted sum
	return framebuffer;
}

//-------------------------------------------------------------------------------------------------
// psSepia
//		Pixel shader used for sepia tone.
//-------------------------------------------------------------------------------------------------
float4 psSepia(float2 tex0 : TEXCOORD0, uniform samplerRECT framebufferSampler, uniform float3 yiqTransform0, uniform float3 yiqTransform1, uniform float3 yiqTransform2, uniform float3 yiqInverseTransform0, uniform float3 yiqInverseTransform1, uniform float3 yiqInverseTransform2) : COLOR
{
	// sample framebuffer texture at current texel
	float4 rgb = texRECT(framebufferSampler, tex0);
	
	// transform to YIQ color space
	float3 yiq;
	yiq.x = dot(yiqTransform0, rgb);
	yiq.y = dot(yiqTransform1, rgb);
	yiq.z = dot(yiqTransform2, rgb);
	
	// set color information to sepia tone
	yiq.yz = float2(0.2, 0);
	
	// inverse transform to RGB color space
	rgb.r = dot(yiqInverseTransform0, yiq);	
	rgb.g = dot(yiqInverseTransform1, yiq);	
	rgb.b = dot(yiqInverseTransform2, yiq);	
	
	// return result in RGB color space
	return rgb;
}

//-------------------------------------------------------------------------------------------------
// red
//		Technique used for the red modulative effect.
//-------------------------------------------------------------------------------------------------
technique red
{
	pass redPass
	{
		FragmentProgram = compile arbfp1 psRed(framebufferSampler);
	}
}

//-------------------------------------------------------------------------------------------------
// blur
//		Technique used for the blur effect, uses 4 rendering passes.
//-------------------------------------------------------------------------------------------------
technique blur
{
	pass blurPass0
	{
		FragmentProgram = compile arbfp1 psBlur(framebufferSampler);
	}
	pass blurPass1
	{
		FragmentProgram = compile arbfp1 psBlur(framebufferSampler);
	}
	pass blurPass2
	{
		FragmentProgram = compile arbfp1 psBlur(framebufferSampler);
	}
	pass blurPass3
	{
		FragmentProgram = compile arbfp1 psBlur(framebufferSampler);
	}
}

//-------------------------------------------------------------------------------------------------
// sharpen
//		Technique used for the sharpen effect, uses 4 rendering passes.
//-------------------------------------------------------------------------------------------------
technique sharpen
{
	pass sharpenPass0
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
	pass sharpenPass1
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
	pass sharpenPass2
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
	pass sharpenPass3
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
}

//-------------------------------------------------------------------------------------------------
// sepia_tone
//		Technique used for the sepia tone conversion effect.
//-------------------------------------------------------------------------------------------------
technique sepia_tone
{
	pass sepiaTonePass
	{
		FragmentProgram = compile arbfp1 psSepia(framebufferSampler, yiqTransform0, yiqTransform1, yiqTransform2, yiqInverseTransform0, yiqInverseTransform1, yiqInverseTransform2);
	}
}

//-------------------------------------------------------------------------------------------------
// sepia_tone_sharpen
//		Technique used to showcase how two effects can be combined, in this case sepia tone 
//		conversion and a sharpening filter to create something similar to an old photograph.
//-------------------------------------------------------------------------------------------------
technique sepia_tone_sharpen
{
	pass sharpenPass0
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
	pass sharpenPass1
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
	pass sharpenPass2
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
	pass sharpenPass3
	{
		FragmentProgram = compile arbfp1 psSharpen(framebufferSampler);
	}
	pass sepiaTonePass
	{
		FragmentProgram = compile arbfp1 psSepia(framebufferSampler, yiqTransform0, yiqTransform1, yiqTransform2, yiqInverseTransform0, yiqInverseTransform1, yiqInverseTransform2);
	}
}


